/*
   Highly optimized, low-level x86 string operations
   -----------------------------------------------------
   Part of AFHRM by Michal Zalewski <lcamtuf@staszic.waw.pl>
   Based on Linux 2.0.xx / i386 kernel code.
*/

#define _I386_STRING_H_
#define _LINUX_STRING_H_


char* strcpy(char * dest,const char *src) {
  __asm__ __volatile__(
    "cld\n"
    "1:\tlodsb\n\t"
    "stosb\n\t"
    "testb %%al,%%al\n\t"
    "jne 1b"
    : /* no output */
    :"S" (src),"D" (dest):"si","di","ax","memory");
  return dest;
}

int strcmp(const char* cs,const char* ct) {
  register int __res;
  __asm__ __volatile__(
    "cld\n"
    "1:\tlodsb\n\t"
    "scasb\n\t"
    "jne 2f\n\t"
    "testb %%al,%%al\n\t"
    "jne 1b\n\t"
    "xorl %%eax,%%eax\n\t"
    "jmp 3f\n"
    "2:\tsbbl %%eax,%%eax\n\t"
    "orb $1,%%eax\n"
    "3:"
    :"=a" (__res):"S" (cs),"D" (ct):"si","di");
  return __res;
}

void* memcpy(void* to, const void* from, size_t n) {
  __asm__ __volatile__(
    "cld\n\t"
    "rep ; movsl\n\t"
    "testb $2,%b1\n\t"
    "je 1f\n\t"
    "movsw\n"
    "1:\ttestb $1,%b1\n\t"
    "je 2f\n\t"
    "movsb\n"
    "2:"
    : /* no output */
    :"c" (n/4), "q" (n),"D" ((long) to),"S" ((long) from)
    : "cx","di","si","memory");
  return (to);
}

size_t strlen(const char * s) {
  register int __res;
  __asm__ __volatile__(
    "cld\n\t"
    "repne\n\t"
    "scasb\n\t"
    "notl %0\n\t"
    "decl %0"
    :"=c" (__res):"D" (s),"a" (0),"0" (0xffffffff):"di");
  return __res;
}

void* memset(void* s, char c,size_t count) {
  __asm__ __volatile__(
    "cld\n\t"
    "rep\n\t"
    "stosb"
    : /* no output */
    :"a" (c),"D" (s),"c" (count)
    :"cx","di","memory");
  return s;
}

void* memmove(void* dest,const void* src, size_t n) {
  if (dest<src)
    __asm__ __volatile__(
      "cld\n\t"
      "rep\n\t"
      "movsb"
      : /* no output */
      :"c" (n),"S" (src),"D" (dest)
      :"cx","si","di");
  else
    __asm__ __volatile__(
      "std\n\t"
      "rep\n\t"
      "movsb\n\t"
      "cld"
      : /* no output */
      :"c" (n),
       "S" (n-1+(const char *)src),
       "D" (n-1+(char *)dest)
      :"cx","si","di","memory");
  return dest;
}

